/* ec_nlms.cxx
 * 
 * Acoustic Echo Cancellation with Normalized Least Means Square Algorithm
 * This implementation tries to fulfill ITU-T G.168 requirements, but without:
 * - Echo Suppression can not be disabled by 2100Hz tone
 * - Echo Cancellation does not detect dial tones (DTMF)
 *
 * 2003, 2004 Copyright, Author: Andre Adrian, Selters, Germany
 * <adrianandre@compuserve.de>
 *
 * This File works together with OpenH323 Version 1.13.4
 */

/* The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * Revision 1.0  2004/06/19 andreadrian
 * First public release
 */
 
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>

/* ================================================================ */

const float ALPHADC = 0.01f;  /* DC Remover Smoothing */

const float M10dB_PCM = 10362.0f;   /* dB values for 16bit PCM */
const float M20dB_PCM = 3277.0f;
const float M25dB_PCM = 1843.0f;
const float M30dB_PCM = 1026.0f;
const float M35dB_PCM = 583.0f;
const float M40dB_PCM = 328.0f;
const float M45dB_PCM = 184.0f;
const float M50dB_PCM = 104.0f;
const float M55dB_PCM = 58.0f;
const float M60dB_PCM = 33.0f;

/* ================================================================ */
/* Exponential Smoothing Lowpass/Highpass */

#define POL   	6     	/* -6dB attenuation per octave per Pol */

class HP8 {
  float lowpass[2*POL+1];
  float highpass[2*POL+1];

public:
  HP8();
  float highpass8(float in);
  float lowpass8(float in);
};

HP8::HP8()
{
  // fprintf(stderr, "HP8");
  memset(this, 0, sizeof(HP8));
}

float HP8::highpass8(float in)
{
  const float AlphaHp = 0.075;  /* controls Transfer Frequence */
  const float Gain8   = 1.45f;	/* POL=6 to undo filter attenuation */

  /* Highpass = Signal - Lowpass. Lowpass = Exponential Smoothing */
  highpass[0] = in;
  int i;
  for (i = 0; i < 2*POL; ++i) {
    lowpass[i+1] += AlphaHp*(highpass[i] - lowpass[i+1]);
    highpass[i+1] = highpass[i] - lowpass[i+1];
  }
  return Gain8*highpass[2*POL];   
}

float HP8::lowpass8(float in)
{
  const float AlphaLp = 0.15f;	/* controls Transfer Frequence */

  /* Lowpass = Exponential Smoothing */
  lowpass[0] = in;
  int i;
  for (i = 0; i < 2*POL; ++i) {
    lowpass[i+1] += AlphaLp*(lowpass[i] - lowpass[i+1]);
  }
  return lowpass[2*POL];   
}

/* ================================================================ */

#define FILTER_LEN  (160*8)
#define FILTER_ADD  (10*8)

/* Vector Dot Product */
double dotp(float a[], float b[]) {
  double suml = 0.0, sumh = 0.0;
  int j;
  
  for (j = 0; j < FILTER_LEN; j+= 2) {
    suml += a[j] * b[j];
    sumh += a[j+1] * b[j+1];
  }
  return suml+sumh;
}


class NLMS {
  float x[FILTER_LEN+FILTER_ADD]; // delayed signal
  float w[FILTER_LEN];	      	  // filter
  int j;      	      	      	  // optimize: less memory copies
  int lastupdate;     	      	  // optimize: iterative dotp(x,x)
  double dotp_x_x;    	      	  // optimize: iterative dotp(x,x)
  
  float s0dc, s1dc;	  // DC-level running average
  HP8 s0hp, s0Lp, s1hp;   // Highpass and Lowpass filters
    
  int update;         	  // the NLMS filter
  int inhibit;        	  // the echo suppression
  
public:
  NLMS();
  float nlms(float mic, float spk, int update);
  void nlms_block(float s0[], float s1[], int taps);
};

NLMS::NLMS()
{
  // fprintf(stderr, "NLMS");
  memset(this, 0, sizeof(NLMS));
  j = FILTER_ADD;
}

/* Normalized Least Mean Square Algorithm 
 * The algorithm was developed by Bernard Widrow
 * book: Widrow/Stearns, Adaptive Signal Processing, Prentice-Hall, 1985
 *
 * The source code is based on Michael Hutson "Acoustic Echo Cancellation
 * using digital signal processing", University of Queensland, 2003,
 * Appendix B.1 nlms_adfilt.c
 *
 * Optimizations for lower CPU load by Andre Adrian.
 *
 * First the filter has to learn the acoustics with given spk and mic signals.
 * After learning the filter can estimate a mic signal from the spk signal.
 * This estimated mic signal is subtracted from the real mic signal
 * and you get an acoustic echo cancellation
 *
 * With update = 1 the filter learns the acoustic out of mic and spk signal
 */
float NLMS::nlms(float mic, float spk, int update)
{
  float d = mic;      	      	      	  // desired signal
  x[j] = spk;

  // calculate output of adaptive filter (estimated mic signal from spk signal)
  float y = dotp(w, x + j);

  // calculate error value (mic signal - estimated mic signal)
  float e = d -y;
  
  if (update) {    
    if (lastupdate) {
      // optimize: iterative dotp(x, x)
      dotp_x_x += (x[j]*x[j] - x[j+FILTER_LEN-1]*x[j+FILTER_LEN-1]);
    } else {
      dotp_x_x = dotp(x+j, x+j);
    } 
      
    // calculate variable step size
    float mikro = 0.5f/dotp_x_x;
    
    // update tap weights (filter learning)
    int i;
    for (i = FILTER_LEN-1; i >= 1; --i) {
      w[i] = w[i] + mikro*e*x[i+j];
    }
  }
  lastupdate = update;
  
  if (--j < 0) {
    // optimize: decrease number of memory copies
    j = FILTER_ADD;
    memmove(x+j+1, x, (FILTER_LEN-1)*sizeof(float));    
  }
  
  return e;
}


/* Block Acoustic Echo Cancellation and Suppression */
void NLMS::nlms_block(float s0[], float s1[], int taps) 
{
  const float J = 32; 	// "max of means" parameter
  
  float s0lpPeak = 0.0f, s1peak = 0.0f;
  float s0lpMaxMean = 0.0f, s1maxMean = 0.0f;
  int i;
  int j = 0;
  for (i = 0; i < taps; ++i) {
    // Mic signal remove DC (highpass filter)
    s0dc += ALPHADC*(s0[i] - s0dc);
    s0[i] -= s0dc;
    
    // Lowpass Filter for Near-End Talk detection
    float s0lp;
    s0lp = s0Lp.lowpass8(s0[i]);
    s0lpMaxMean += fabsf(s0lp);
    
    // Highpass Filter
    s0[i] = s0hp.highpass8(s0[i]);

    // Spk signal remove DC (highpass filter)
    s1dc += ALPHADC*(s1[i] - s1dc);
    s1[i] -= s1dc;
    s1maxMean += fabsf(s1[i]);
    
    // for fast detection and robustness use "max of means"
    if (++j >= J) {
      j = 0;
      s0lpMaxMean /= J;
      s1maxMean /= J;
      
      if (s0lpMaxMean > s0lpPeak) {
	s0lpPeak = s0lpMaxMean;
      }
      if (s1maxMean > s1peak) {
	s1peak = s1maxMean;
      }
      s0lpMaxMean = 0.0f;
      s1maxMean = 0.0f;
    }
  }
  
  /* Detect state (silence, single listen, single talk, double talk)
   * This is the non-mathematical part of acoustic echo cancellation:
   * determine what the NLMS should do.
   * The software is tuned for good double talk performance in small to
   * medium-sized rooms with not much ambient noise.
   */
  int suppress = 0;   	    // the Mic signal
  char state = ' ';
  
  if (s0lpPeak > M45dB_PCM) { 	// near-end single talk or double talk
    if (s1peak > M45dB_PCM) { 	// double talk
      inhibit = 500*8 / taps; 	// silence within one word
      update = 0;
      state = 'd';
     } else {  // single talk
      update = 0;
      state = 't';
    }
  } else {    // silence or listen
    if (s1peak > M25dB_PCM) { 	// listen
      update = FILTER_LEN / taps;
      suppress = 1;
      state = 'l';
    } else {  	// silence
      state = 's';
    }
  }
  if (update) {
    state = toupper(state);
  }
  // fprintf(stderr, "%c", state);
  
  // Acoustic Echo Cancellation
  for (i = 0; i < taps; ++i) {
    s0[i] = nlms(s0[i], s1[i], update);
  }

  // Acoustic Echo Suppression
  if (suppress && !inhibit) {
    // Non Linear Processor (NLP): attenuate low volumes
    for (i = 0; i < taps; ++i) {
      s0[i] *= 0.5f;
    }
  }
  
  // work the timers
  if (inhibit) --inhibit;
  if (update) --update;
}

typedef signed short MONO;    /* mono signal: assumption 16-bit format */
typedef struct {
  signed short	l;
  signed short	r;
} STEREO;     	      	      /* stereo signal: assumption 16-bit format */

#define TAPS (80*8)

/* ================================================================ */

/* copied from openh323/src/codecs.cxx */

#include <ptlib.h>

#ifdef __GNUC__
#pragma implementation "codecs.h"
#endif

#include "codecs.h"

#include "channels.h"
#include "h323pdu.h"
#include "h323con.h"


#define new PNEW

#if !PTRACING // Stuff to remove unised parameters warning
#define PTRACE_bitRateRestriction
#define PTRACE_type
#endif

extern "C" {
  unsigned char linear2ulaw(int pcm_val);
  int ulaw2linear(unsigned char u_val);
  unsigned char linear2alaw(int pcm_val);
  int alaw2linear(unsigned char u_val);
};

/* copied from openh323/src/codecs.cxx H323Codec::ReadRaw() */
/* Attention: Subclassing will not work, we have to overwrite the original
 * methods */
BOOL H323Codec::ReadRaw(void * data, PINDEX size, PINDEX & length)
{
  if (rawDataChannel == NULL) {
    PTRACE(1, "Codec\tNo audio channel for read");
    return FALSE;
  }

  /* adrian: double buffer size because we have stereo */
  STEREO sdata[TAPS]; 
  if (!rawDataChannel->Read(sdata, 2*size)) {
    PTRACE(1, "Codec\tAudio read failed: " << rawDataChannel->GetErrorText(PChannel::LastReadError));
    return FALSE;
  }

  /* adrian: do NLMS for all taps */
  MONO *mdata = (MONO *)data;
  int taps = size/sizeof(MONO);     // how many taps in one audio packet
  float s0[TAPS], s1[TAPS];
  for (int i = 0; i < taps; ++i) {
    s0[i] = (float)sdata[i].l;
    s1[i] = (float)sdata[i].r;
  }
  
  static NLMS n0;
  n0.nlms_block(s0, s1, taps);

  /* copy back */
  for (int i = 0; i < taps; ++i) {
    mdata[i] = (MONO)(s0[i]);
  }
  
  length = rawDataChannel->GetLastReadCount();

  /* adrian: stereo length to mono length */
  length /= 2;
  
  for (PINDEX i = 0; i < filters.GetSize(); i++) {
    FilterInfo info(*this, data, size, length);
    filters[i](info, 0);
    length = info.bufferLength;
  }
  return TRUE;
}


/* copied from openh323/src/codecs.cxx H323Codec::WriteRaw() */
BOOL H323Codec::WriteRaw(void * data, PINDEX length)
{
  if (rawDataChannel == NULL) {
    PTRACE(1, "Codec\tNo audio channel for write");
    return FALSE;
  }
  
  for (PINDEX i = 0; i < filters.GetSize(); i++) {
    FilterInfo info(*this, data, length, length);
    filters[i](info, 0);
    length = info.bufferLength;
  }

  /* mono to stereo */
  STEREO sdata[TAPS];
  MONO *mdata = (MONO *)data;
  int taps = length/sizeof(MONO);     // how many taps in one audio packet
  for (int i = 0; i < taps; ++i) {
    sdata[i].l = 0;
    sdata[i].r = mdata[i];
  }
  
  if (rawDataChannel->Write(sdata, 2*length))
    return TRUE;

  PTRACE(1, "Codec\tWrite failed: " << rawDataChannel->GetErrorText(PChannel::LastWriteError));
  return FALSE;
}
